@@ -9,9 +9,9 @@ import inspect
 import vxi11
 import time
 #===========================================================================
-# Oscilloscope DS1102Z_E,  Pete (v2.6 tested)
+# Oscilloscope - Rigol DS1000E series 
 #===========================================================================            
-class Rigol_DS1102Z_E:
+class Rigol_DS1000E:
     def __init__(self,address): # 'TCPIP::<IP ADDRESS>::INSTR'
         self.address = address
         self.rm = pyvisa.ResourceManager()
@@ -59,9 +59,9 @@ class Rigol_DS1102Z_E:
             self.dev.write(':TRIG:EDGe:LEV {}'.format(__vTrig))  
             
 #================================================================
-# Rigol_DSA832 Spectrum Analyzer ()
+# Spectrum Analyzer - Rigol_DSA800 series 
 #================================================================
-class Rigol_DSA832:
+class Rigol_DSA800:
     def __init__(self,address): #'TCPIP::<IP ADDRESS>::INSTR' 
         self.address = address            
         self.rm = pyvisa.ResourceManager()
@@ -216,33 +216,22 @@ class Rigol_DS1104:
         self.dev.write(':SINGle')
     
     def screenshot(self,ch):
-        
-        timescale = float(self.dev.query(":TIM:SCAL?"))        
-        timeoffset = float(self.dev.query(":TIM:OFFS?")[0])# Get the timescale offset
-        voltscale = float(self.dev.query(":CHAN{}:SCAL?".format(ch)))
-        voltoffset = float(str(self.dev.query(":CHAN{}:OFFS?".format(ch))))  # And the voltage offset
-        
-        self.dev.write(":WAV:SOUR: CHAN{}".format(ch))
-        self.dev.write(":WAV:POIN:MODE NORM")
+        self.dev.write(":WAV:SOUR CHAN{}".format(ch))
+        self.dev.write(":WAV:MODE NORMal")
+        self.dev.write(":WAV:FORM ASC")
         self.dev.write(":WAV:DATA? CHAN{}".format(ch))
-        rawdata = self.dev.read_raw() #1024 data
-        rawdata = rawdata[11:]
-
-        data = np.frombuffer(rawdata, 'B')
+        rawdata = self.dev.read_raw()
+        rawdata = rawdata.decode('UTF-8')
+        volt_data = rawdata[11:-2] #removes header and ending of data
+        volt_data = np.array([float(volt_data) for volt_data in volt_data.split(',')])
         
-        # Walk through the data, and map it to actual voltages
-        data = data + 254
-        data = data[:-1]
-
-        # Now, we know from experimentation that the scope display range is actually
-        # 30-229.  So shift by 130 - the voltage offset in counts, then scale to
-        # get the actual voltage.
-        data = (data - 130.0 - voltoffset/voltscale*25) / 25 * voltscale
+        t = float(self.dev.query(':WAVeform:XINCrement?'))
+        time_data = np.arange(0,t*len(volt_data),t)
         
-        return data
+        return time_data,volt_data
 
 #===========================================================================
-# Oscilloscope Rigol DMO5000
+# Oscilloscope - Rigol DMO5000
 #=========================================================================== 
 
 class Rigol_DMO5000:
@@ -269,7 +258,7 @@ class Rigol_DMO5000:
         self.dev.write(':SINGle')
         
     def screenshot(self,ch):
-        self.dev.write(":WAV:SOUR: CHAN{}".format(ch))
+        self.dev.write(":WAV:SOUR CHAN{}".format(ch))
         self.dev.write(":WAV:MODE NORMal")
         self.dev.write(":WAV:FORM ASC")
         self.dev.write(":WAV:DATA? CHAN{}".format(ch))
@@ -281,4 +270,124 @@ class Rigol_DMO5000:
         t = float(self.dev.query(':WAVeform:XINCrement?'))
         time_data = np.arange(0,t*len(volt_data),t)
         
-        return time_data,volt_data
\ No newline at end of file
+        return time_data,volt_data
+    
+#===========================================================================
+# Polarimeter - Thorlabs PAX1000
+#=========================================================================== 
+
+class thorlabs_polarimeter:
+    def __init__(self,address):
+        """ 
+        Description: The initialization of the polarimeter class 
+        will initialize a singular device using its VISA USB address
+
+        Input: VISA address : string
+        Output: device object with all its methods : class object
+        Example: 
+        >>thorlabs_polarimeter('USB::0x1234::125::A22-5::INSTR')
+
+        thorlabs_polarimeter.class.object...
+        """
+        self.address = address
+        self.rm = pyvisa.ResourceManager()
+        self.dev = self.rm.open_resource(self.address)
+        self.dev.write('SENS:CALC:MODE 1')
+        self.dev.write('SENS:POW:RANG:IND 8')
+        
+    def set_wavelength(self,wavelength):
+        """ 
+        Description: sets wavelength
+
+        Input: wavelength : float
+        Output: None : None
+        Example: 
+        >>set_wavelength(795)
+
+        """
+        if bool(re.search('PAX1000IR1', self.dev.query('*IDN?'))):
+            if wavelength < 1080 and wavelength > 600:
+                self.dev.write('SENS:CORR:WAV {}'.format(wavelength*1e-9))
+            else:
+                raise ValueError('Wavelength not supported (600-1080nm)')
+        elif bool(re.search('PAX1000IR2', self.dev.query('*IDN?'))):
+            if wavelength < 1700 and wavelength > 900:
+                self.dev.write('SENS:CORR:WAV {}'.format(wavelength*1e-9))
+            else:
+                raise ValueError('Wavelength not supported (900-1700nm)')
+        else:
+            print('check polarimeter connection')
+        return
+
+    def trigger_data_collection(self):
+        """ 
+        Description: opens data collection window
+
+        Input: None : None
+        Output: None : None
+        Example: 
+        >>trigger_data_collection()
+
+        """
+        self.dev.write('INP:ROT:STAT 1')
+        return
+        
+        
+    def close_data_collection(self):
+        """ 
+        Description: closes data collection window
+
+        Input: None : None
+        Output: None : None
+        Example: 
+        >>close_data_collection()
+
+        """
+        self.dev.write('INP:ROT:STAT 0')
+        return
+    
+    def get_polarzation_params(self):
+        """ 
+        Description: Returns the current measured Stokes parameters
+        for the polarimeter
+
+        Input: None : None
+        Output: Stokes Parameters : tuple
+        Example: 
+        >>get_polarzation_params()
+
+        20, 28, 19, 1, 1
+        """
+        data = self.dev.query('SENS:DATA:LAT?')
+        data = data.split(',')
+        psi = float(data[9])
+        chi = float(data[10])
+        dop= float(data[11])
+        power = float(data[12])
+        s1 = np.cos(2 * psi) * np.cos(2 * chi)
+        s2 = np.sin(2 * psi) * np.cos(2 * chi)
+        s3 = np.sin(2 * chi)
+        return s1, s2, s3, dop, power
+
+    def get_raw_data(self):
+        """ 
+        Description: Returns the raw data read from the polarimeter 
+
+        Input: None : None
+        Output: Raw Data from Polarimeter : string
+        """
+        data = self.dev.query('SENS:DATA:LAT?')
+        return data
+    
+    def reset(self):
+        """ 
+        Description: resets connection with polarimeter
+
+        Input: None : None
+        Output: Reset of connection : class method
+        """
+        self.dev.close()
+        time.sleep(1)
+        self.dev = self.rm.open_resource(self.address)
+        self.dev.write('*RST')
+        
\ No newline at end of file
