@@ -65,10 +65,11 @@ class Valon_5015:
 #         self.dev.close()
 
 #================================================================
-# Rigol_DSG830 (tested)
+# Function Generator - Rigol DSG800 series
 #================================================================
 
-class Rigol_DSG830:
+class Rigol_DSG800:
+    
     def __init__(self,address,*arg):
         if arg:
             self.address = address #'TCPIP::<IP ADDRESS>::INSTR'
@@ -80,6 +81,7 @@ class Rigol_DSG830:
      
     def off(self):
         self.dev.write(':OUTput OFF;')  # turn on RF output
+        
     def on(self):
         self.dev.write(':OUTput On;')  # turn on RF output
     
@@ -112,17 +114,13 @@ class Rigol_DSG830:
             return __lev
 
 #================================================================
-# Rigol_DG4202 (1.2 tested)
+# Function Generator - Rigol DG4000 series
 #================================================================
-class Rigol_DG4202:
-    def __init__(self,address,*arg):
-        if arg:
-            self.address = address #'TCPIP::<IP ADDRESS>::INSTR'
-            self.dev = vxi11.Instrument(self.address)
-        else:
-            self.address = address 
-            self.rm = pyvisa.ResourceManager()
-            self.dev = self.rm.open_resource(self.address)
+class Rigol_DG4000:
+    def __init__(self,address):
+        self.address = address
+        self.rm = pyvisa.ResourceManager()
+        self.dev = self.rm.open_resource(self.address)
         
     def reset(self): #Reset
         self.dev.write("*RST")
@@ -142,10 +140,14 @@ class Rigol_DG4202:
             self.dev.write(':OUTput1 ON') 
             self.dev.write(':OUTput2 ON')
 
-    def freq(self,ch,f):
-        self.dev.write(':SOURCe{}:Freq {}'.format(ch,self.__Hz(f)));
+    def __Hz(f):  # in Hz, support unit. Default: MHz
+        """
+        Description: Sets all frequencies to MHz Unit (Tested 04/03/2022)
 
-    def __Hz(self,f):  # in Hz, support unit. Default: MHz
+        Input: f : frequency : float
+
+        Output: f * 1e6 : frequency in MHz : float
+        """
         if type(f) == str:
             if re.search('[mM]',f) != None:
                 return 1e6*float(re.sub('[a-zA-Z]','',f))
@@ -156,22 +158,43 @@ class Rigol_DG4202:
         else: # float, or str that contains only numbers
             return float(f)*1e6
 
-    def lev(self,ch,v):
-        if type(v) == str:
-            __V = float(re.sub('[a-zA-Z]','',v))  # unitless value
-            if re.search('[rR]',v) != None:  #  VRMS
-                self.dev.write(':SOURCe{}:VOLTage:UNIT VRMS'.format(ch))
-            elif re.search('[dD]',v) != None:  # dBm
-                self.dev.write(':SOURCe{}:VOLTage:UNIT DBM'.format(ch))
-        else:  # default: [Vpp] 
-            __V = v
-            self.dev.write(':SOURCe{}:VOLTage:UNIT VPP'.format(ch))
+    def freq(self,ch,*f):
+        if bool(f) == True:   # assign value (only the first in the tuple)
+            __freq = f[0]
+            self.dev.write(':SOURCe{}:Freq {}'.format(ch,self.__Hz(__freq)));
+        else:   # when no input is entered, read actual frequency 
+            __readOut = self.dev.query(':SOURCe{}:Freq?'.format(ch))
+            __freq = float(re.search('.*(?=\n)',__readOut).group())
+            return float(__freq)*1e-6  # MHz
+        
+    def lev(self,ch,*v):
+        if bool(v) == True:
+            __v = v[0]
+            if type(__v) == str:
+                __lev = float(re.sub('[a-zA-Z]','',__v))  # unitless value
+                if re.search('(vrms)',__v,re.IGNORECASE) != None:  #  VRMS
+                    self.dev.write(':SOURCe{}:VOLTage:UNIT VRMS'.format(ch))
+                elif re.search('d',__v,re.IGNORECASE) != None:  # dBm
+                    self.dev.write(':SOURCe{}:VOLTage:UNIT DBM'.format(ch))
+                else:  # VPP
+                    self.dev.write(':SOURCe{}:VOLTage:UNIT VPP'.format(ch))
+                # mVPP or mVRMS
+                if re.search('(mv)',__v,re.IGNORECASE) != None:
+                    __lev = 1e-3 * __lev
+                # value
+                self.dev.write(':SOURCe{}:VOLTage {}'.format(ch,__lev))
+            else:  # default: [Vpp] 
+                self.dev.write(':SOURCe{}:VOLTage:UNIT VPP'.format(ch))
+                self.dev.write(':SOURCe{}:VOLTage {}'.format(ch,__v))
+        else:
+            __readOut = self.dev.query(':SOURCe{}:VOLTage?'.format(ch))
+            __lev = float(re.search('.*(?=\n)',__readOut).group())
+            __readOut = self.dev.query(':SOURCe{}:VOLTage:UNIT?'.format(ch))
+            __unit = re.search('.*(?=\n)',__readOut).group()
+            return __lev, __unit
 
-        self.dev.write(':SOURCe{}:VOLTage {}'.format(ch,__V))
-        self.dev.write(':SOURCe{}:VOLTage:UNIT VPP'.format(ch))
-
-    def offset(self,ch,o):  # V_DC
-        self.dev.write(':SOURCe{}:VOLTage:OFFSet {}'.format(ch,o));
+    def offset(self,ch,offset):  # V_DC
+        self.dev.write(':SOURCe{}:VOLTage:OFFSet {}'.format(ch,offset));
 
     def phase(self,ch,p):
         self.dev.write(':SOURCe{}:PHASe {}'.format(ch,p));
@@ -184,61 +207,120 @@ class Rigol_DG4202:
     def square(self,t,leadingedge,width,amp): #square pulse with duty cycle
         return np.piecewise(t,[(t<=leadingedge),((t>leadingedge) & (t<leadingedge+width)),(t>=leadingedge+width)],[0,amp,0])
     
-    def arb(self,ch,ν,func,*arg):
-        total_time = 1/(self.__Hz(ν))
+    def normalize(self,waveform):
+        """
+        Description: Normalizes data for arbitrary waveform design, points are limited (Tested 04/03/2022)
+        to -1 to 1 Volt
+
+        Input: data : waveform : np.array or list
+
+        Output: np.array(waveform)/np.absolute(max(waveform)) : normalized data :  np.array or list
+        """
+        return np.array(waveform)/np.absolute(max(waveform))
         
-        if inspect.ismethod(func) == True:
-            t = np.linspace(0,total_time,1000)
-            data = func(t,*arg)
-            datastring = ",".join(map(str,data))
+    def arbitrary(self, ch, signal_width, waveform, *arg):
+        """
+        Description: Allows one to set and create arbitrary waveform output (Tested 04/03/2022) 
+
+        Input: ch : channel : int
+               signal_width : width (time) of the argument in seconds : float
+               waveform : method or list of values : method or array
+               arg* : arguments of the waveform function : misc
+
+        Output: None : class method
+        """
+        buffer_size = 2**14
+        if inspect.ismethod(waveform) == True:
+            t = np.linspace(0,signal_width,buffer_size)
+            data = np.around(waveform(t,*arg),4)
+            datastring = ",".join(map(str,self.normalize(data)))
         else:
-            data = func
-            datastring = ",".join(map(str,data))
-        
-        self.dev.write("SOURCE{}:BURST OFF".format(ch))
-        self.dev.write("OUTPUT{} ON".format(ch))
-        self.dev.write("SOURCE{}:TRACE:DATA VOLATILE,".format(ch)+ datastring)
-        self.dev.write("SOURCE{}:VOLTAGE:UNIT VRMS".format(ch))
-        self.dev.write('SOURCE{}:Freq {}'.format(ch,self.__Hz(ν)))
-        self.dev.write("SOURCE{}:VOLTAGE:LOW {}".format(ch,min(data)))
-        self.dev.write("SOURCE{}:VOLTAGE:HIGH {}".format(ch,max(data)*2))
-        self.dev.write("SOURCE{}:VOLTAGE:OFFSET 0".format(ch))
-        self.dev.write("SOURCE{}:PHASE 0".format(ch))
-        self.dev.write("SOURCE{}:PERIOD {}".format(ch,total_time))
+            data = np.around(waveform,4)
+            datastring = ",".join(map(str,self.normalize(data)))
+
+        self.dev.write('SOURCE{}:Freq {}'.format(ch, 1/signal_width))
+        self.dev.write("SOURCE{}:TRACE:DATA VOLATILE,".format(ch) + datastring)
+        self.dev.write("SOURCE{}:FUNC VOLATILE,".format(ch))
+        if len(data[data < 0]) == 0:
+            self.dev.write("SOURCE{}:VOLTAGE:LOW {}".format(ch,min(data)-(max(data)-min(data))))
+            self.dev.write("SOURCE{}:VOLTAGE:HIGH {}".format(ch,min(data)+(max(data)-min(data))))
+        else: 
+            self.dev.write("SOURCE{}:VOLTAGE:LOW {}".format(ch,min(data)))
+            self.dev.write("SOURCE{}:VOLTAGE:HIGH {}".format(ch,max(data)))
         self.dev.write("SOURCE{}:PHASE:SYNC".format(ch))
     
+    def burst(self, ch, mode, cycles):
+        """
+        Description: Allows on to use burst functionallity (Tested 04/03/2022)
+
+        Input: ch : channel : int
+               mode : burst mode : int = {'TRIG' = 0, 'GAT' = 1,'INF' = 2}
+               cycles : cycles : int
+
+        Output: None : class method
+        """
+        modes = ['TRIGgered','GATed','INFinity']
+        self.dev.write('SOURce{}:BURSt:MODE {}'.format(ch,modes[mode]))
+        self.dev.write('SOURce{}:BURSt:TRIGger:SOURce EXT'.format(ch))
+        self.dev.write('SOURce{}:BURSt:NCYCles {}'.format(ch,cycles))
+        self.dev.write('SOURce{}:BURSt ON'.format(ch))
+        
+    def DC(self, ch, offset):
+        """
+        Description: Enables DC Mode (Tested 04/03/2022)
+
+        Input: ch : channel : int
+               offset : DC Offset : float
+
+        Output: None : class method
+        """
+        self.dev.write("SOURCE{}:FUNC DC".format(ch))
+        self.offset(ch,offset)
+    
     def ext_trig(self,ch):
         self.dev.write("SOURCE{}:BURST:TRIG:SOUR EXT".format(ch))
-    
-    def arb_burst(self,ch,ν,cycles,func,*arg):
-        total_time = 1/(self.__Hz(ν))
         
+    def arbitrary_burst(self,ch,signal_width,cycles,func,*arg):
+        """
+        Description: Uses N cycle burst functionallity of the Rigol DG4000 Series. Best 
+        suited for external trigger.
+
+        Input: ch : channel : int
+               signal_width : width (time) of the argument in seconds : float
+               cycles : cycles : int
+               func : method or list of values : method or array
+               arg* : arguments of the function : misc
+        Output: None : class method
+        """
         if inspect.ismethod(func) == True:
-            t = np.linspace(0,total_time,1000)
+            t = np.linspace(0,signal_width,1000)
             data = func(t,*arg)
-            datastring = ",".join(map(str,data))
+            datastring = ",".join(map(str, self.normalize(data)))
         else:
             data = func
-            datastring = ",".join(map(str,data))
-        
+            datastring = ",".join(map(str, self.normalize(data)))
+
+        self.dev.write('SOURCE{}:Freq {}'.format(ch, 1/signal_width))
         self.dev.write("OUTPUT{} ON".format(ch))
         self.dev.write("SOURCE{}:TRACE:DATA VOLATILE,".format(ch)+ datastring)
-        self.dev.write("SOURCE{}:VOLTAGE:UNIT VRMS".format(ch))
-        self.dev.write('SOURCE{}:Freq {}'.format(ch,self.__Hz(ν)))
-        self.dev.write("SOURCE{}:VOLTAGE:LOW {}".format(ch,min(data)))
-        self.dev.write("SOURCE{}:VOLTAGE:HIGH {}".format(ch,max(data)*2))
+        self.dev.write("SOURCE{}:VOLTAGE:UNIT VPP".format(ch))
+        self.dev.write("SOURCE{}:VOLTAGE:AMPL {}".format(ch,2*max(data)))
+        # self.dev.write("SOURCE{}:VOLTAGE:LOW {}".format(ch,min(data)))
+        # self.dev.write("SOURCE{}:VOLTAGE:HIGH {}".format(ch,max(data)))
         self.dev.write("SOURCE{}:VOLTAGE:OFFSET 0".format(ch))
         self.dev.write("SOURCE{}:PHASE 0".format(ch))
-        self.dev.write("SOURCE{}:PERIOD {}".format(ch,total_time))
         self.dev.write("SOURCE{}:PHASE:SYNC".format(ch))
-        
+
         #triggered burst
         self.dev.write("SOURCE{}:BURST ON".format(ch))
         self.dev.write("SOURCE{}:BURST:NCYC {}".format(ch,cycles))
         self.dev.write("SOURCE{}:BURST:MODE:TRIG".format(ch))
+        self.ext_trig(ch)
+
+        
         
 #===========================================================================
-#MOGLabs (1.0 tested)
+#MOGLabs
 #===========================================================================
 class MOGDevice:
     def __init__(self,addr,port=None,timeout=1,check=True):
@@ -432,10 +514,7 @@ class MOGDevice:
             if val is not None: self.dev.settimeout(val)
             return old
 
-#===========================================================================
-# MOGLabs
-#===========================================================================         
-
+# Driver
 class MOGLabs:
     def __init__(self,address): #<IP ADDRESS>
         self.address = address
@@ -454,17 +533,17 @@ class MOGLabs:
         else:
             self.dev.cmd('On,1')
 
-    def freq(self,ch,f):
+    def freq(self,ch,frequency):
         if type(f) == str:
-            self.dev.cmd('FREQ,{},{}'.format(ch,f))
+            self.dev.cmd('FREQ,{},{}'.format(ch,frequency))
         else:
-            self.dev.cmd('FREQ,{},{} MHz'.format(ch,f))
+            self.dev.cmd('FREQ,{},{} MHz'.format(ch,frequency))
 
-    def lev(self,ch,f):
-        if type(f) == str:
-            self.dev.cmd('POW,{},{}'.format(ch,f))
+    def lev(self,ch,amplitude):
+        if type(amplitude) == str:
+            self.dev.cmd('POW,{},{}'.format(ch,amplitude))
         else:
-            self.dev.cmd('POW,{},{} dBm'.format(ch,f))
+            self.dev.cmd('POW,{},{} dBm'.format(ch,amplitude))
 
     def am(self,ch=1):  # turn on amplitude modulation
         self.on(ch)    # turn on RF output
@@ -479,8 +558,9 @@ class MOGLabs:
         self.dev.cmd('POW,{},30 dBm'.format(ch) )
         
 #===========================================================================
-# Quantum_Composer (1.0 tested)
+# Quantum Composers
 #===========================================================================
+
 def sleep_method(method, *arg):
     t_sleep = 50e-3
     def sleeping_method(self, *arg):
@@ -508,6 +588,7 @@ class Quantum_Composers:
             self.mux_reset()   # clear all multiplexer
             self.lev()    # set all outputs to TTL
             self.t_sleep = 50e-3
+            self.digit   = 11   # important! round evertying to 11 digits
         else:
             self.address = address
             self.rm = pyvisa.ResourceManager()
@@ -516,41 +597,63 @@ class Quantum_Composers:
             self.lev()    # set all outputs to TTL
             self.dev.clear()
             self.t_sleep = 50e-3
+            self.digit   = 11   # important! round evertying to 11 digits
 
+    def rd(self,x,*digit):
+        if bool(digit) == False:
+            return round(x,self.digit)
+        else:
+            return round(x,digit)
+        
     @sleep_method
     def t0(self,t):  # clock T0
+        t = self.rd(t)
         self.dev.write(':PULSE0:PER {}'.format(t))
+        
     @sleep_method
     def norm(self,*ch):   # normal mode, no wait
         if bool(ch) == True:  # specified channel
             for __ch in ch:
-                self.dev.write(':Pulse{}:CMODe NORMal'.format(__ch))
+                if __ch > 0: # channel model
+                    self.dev.write(':Pulse{}:CMODe NORMal'.format(__ch))
+                elif __ch == 0:  # T0 mode
+                    self.dev.write(':Pulse0:MODe NORMal')
         else:
-            for __ch in range(1,9):  # all channels
+            for __ch in range(1,9):  # all channels and T0
                 self.dev.write(':Pulse{}:CMODe NORMal'.format(__ch))
+                self.dev.write(':Pulse0:MODe NORMal')
+                
     @sleep_method            
     def wid(self,ch,w):
+        w = self.rd(w)
         self.dev.write(':PULSE{}:WIDth {}'.format(ch,w))
+        
     @sleep_method
     def dly(self,ch,d):
+        d = self.rd(d)
         self.dev.write(':PULSE{}:DELay {}'.format(ch,d))
+        
     @sleep_method
     def pol(self,ch,p):
         self.dev.write(':PULSE{}:POL {}'.format(ch,p))
+        
     @sleep_method
-    def wcount(self,ch,w):  # wait number of T0 before enable output
+    def wcount(self,ch,w):  # wait number of T0 before enable output       
         self.dev.write(':PULSE{}:WCOunter {}'.format(ch,w))
+        
     @sleep_method
     def dcycl(self,ch,on,off):   # channel duty cycle
         self.dev.write(':Pulse{}:CMODe DCYCLe'.format(ch))
         self.dev.write(':PULSE{}:PCOunter {}'.format(ch,on))
         self.dev.write(':PULSE{}:OCOunter {}'.format(ch,off))
+        
     @sleep_method
     def lev(self,*p):    # set the output amplitude of a channel
         if bool(p) == True:  # adjustbale
             __ch = p[0]
             if len(p) > 1:
                 __volt = p[1]
+                __volt = self.rd(__volt)
                 self.dev.write(':PULSE{}:OUTPut:MODe ADJustable'.format(__ch))
                 self.dev.write(':PULSE{}:OUTPut:AMPLitude {}'.format(__ch,__volt))
             else:
@@ -558,6 +661,7 @@ class Quantum_Composers:
         else:   # TTL
             for __ch in range(1,9): 
                 self.lev(__ch)
+                
     @sleep_method
     def mux(self,*p):  ## multiplexer
         if bool(p) == True:
@@ -569,10 +673,12 @@ class Quantum_Composers:
             self.dev.write(':PULSE{}:MUX {}'.format(__ch,__binary))
         else:
             self.mux_reset()
+            
     @sleep_method
     def mux_reset(self):   # reset multiplexer
         for n in range(1,9): 
             self.mux(n,n)
+            
     @sleep_method
     def on(self,*ch):
         if bool(ch) == True:
@@ -580,6 +686,7 @@ class Quantum_Composers:
                 self.dev.write(':PULSE{}:STAT ON'.format(channel))
         else:
             self.dev.write(':PULSE0:STAT ON')
+            
     @sleep_method
     def off(self,*ch):
         if bool(ch) == True:
@@ -593,6 +700,10 @@ class Quantum_Composers:
         self.dev.write(':PULSE0:TRIG:MOD TRIG')  # trig enabled 
         self.on()
         self.dev.write('*TRG')  # software trigger
+
+    @sleep_method        
+    def cw(self): # continuous running mode
+        self.dev.write(':PULSE0:TRIG:MOD DIS')  # trig disabled 
         
     @sleep_method
     def trigOff(self):
@@ -603,13 +714,19 @@ class Quantum_Composers:
         for c in ch:
             self.dev.write(':PULSE{}:POL INV'.format(c))
             self.off(c)
+            
     @sleep_method
     def low(self,*ch):  ## keep output constantly at 0V
         for c in ch:
             self.dev.write(':PULSE{}:POL NORM'.format(c))
             self.off(c)
+            
     @sleep_method   
     def __exp(self,T0,ch,pol,tExp,tPls,tDly,nPls,nDly):  # experiment mode
+        T0 = self.rd(T0)
+        tExp = self.rd(tExp)
+        tPls = self.rd(tPls)
+        tDly = self.rd(tDly)
         __dcycl_on = nPls
         __dcycl_tot = round(tExp/T0)
         __dcycl_off = __dcycl_tot - __dcycl_on
@@ -619,6 +736,7 @@ class Quantum_Composers:
         self.dcycl(ch,__dcycl_on,__dcycl_off)
         self.wcount(ch,nDly)  
         self.on(ch)
+        
     @sleep_method    
     def config(self,cfg):   # preset configuration
         if re.search('(cal)', cfg, re.IGNORECASE)!= None:   # cavity calibration
@@ -632,98 +750,10 @@ class Quantum_Composers:
         elif re.search('(off)', cfg, re.IGNORECASE)!= None:  # all off
             self.mux_reset()
             self.off(0,1,2,3,4,5,6,7,8)
-        elif re.search('c.*(eit)', cfg,re.IGNORECASE) != None:
-            self.mux_reset()
-            self.off(1,2,3,4,5,6,7,8,0)
-            self.trigOff() ## setup system
-            __T0 = 100e-6
-            __tExp = 3.3
-            __tResp = 0e-6  # time that cooling may remain ~0.2 us
-            self.dev.write(':PULSE0:PER {}'.format(__T0))  # 0.000100080   
-            ## Channel A: Cooling & Magnetic Field 
-            self.__exp(__T0,1,'NORM',__tExp,0.3,0.0,1,30000) 
-            ## Channel B: Repumper
-            self.__exp(__T0,2,'INV',__tExp,0.3,0.0,1,30000)
-            ## Channel C: Probe
-            self.__exp(__T0,3,'NORM',__tExp,0.1,__tResp,1,30000)
-            ## Channel D: Control
-            self.__exp(__T0,4,'NORM',__tExp,0.000050,__tResp,1000,30000)
-            ## Channel H: TTL for quTAU
-            self.__exp(__T0,8,'NORM',__tExp,0.000050,__tResp,1000,30000)
-            self.on(1,2,3,4,8)   # this won't start the trigger mode
-        ## DLCZ_1, d2 photon generation
-        elif (re.search('(photon)|(d2)', cfg ,re.IGNORECASE) != None) and (re.search('(no wait)', cfg ,re.IGNORECASE) == None):  
-            self.mux_reset()
-            self.off(1,2,3,4,5,6,7,8,0)
-            self.trigOff() ## setup system
-#            __T0 = 100e-6
-            __T0 = 10e-6
-            __tExp = 3.3
-            self.dev.write(':PULSE0:PER {}'.format(__T0))  # 0.000100080   
-            ## Channel A: Cooling & Magnetic Field 
-            self.__exp(__T0,1,'INV',__tExp,3.0,0.0,1,0) 
-            ## Channel B: Repumper
-            self.__exp(__T0,2,'NORM',__tExp,3.0,0.0,1,0)
-            ## Channel D: Pump
-            self.__exp(__T0,4,'NORM',__tExp,0.000005,0.0,10000,300001)
-#             self.__exp(__T0,4,'NORM',__tExp,0.0000025,0.0,10000,300001)
-            ## Channel H: TTL for quTAU
-            self.__exp(__T0,8,'NORM',__tExp,0.000005,0.0,10000,300000) 
-#             self.__exp(__T0,8,'NORM',__tExp,0.0000025,0.0,10000,300000)
-            self.on(1,2,4,8)
-        elif re.search('(d2 photon no wait)',cfg,re.IGNORECASE) != None:  
-            self.mux_reset()
-            self.off(1,2,3,4,5,6,7,8,0)
-            self.trigOff() ## setup system
-            __T0 = 100e-6
-            __tExp = 3.3
-            __tResp = 0e-6  # no waiting 
-            self.dev.write(':PULSE0:PER {}'.format(__T0))  # 0.000100080   
-            ## Channel A: Cooling & Magnetic Field 
-            self.__exp(__T0,1,'INV',__tExp,3.0,0.0,1,0) 
-            ## Channel B: Repumper
-            self.__exp(__T0,2,'NORM',__tExp,3.0,0.0,1,0)
-            ## Channel D: Pump
-            self.__exp(__T0,4,'NORM',__tExp,0.000050,__tResp,1000,30000)
-            ## Channel H: TTL for quTAU
-            self.__exp(__T0,8,'NORM',__tExp,0.000050,__tResp,1000,30000)
-            self.on(1,2,4,8)
-        elif re.search('(fluo)',cfg,re.IGNORECASE) != None:    # fluorescence
-            self.mux_reset()
-            self.off(1,2,3,4,5,6,7,8,0)
-            self.trigOff() ## setup system
-            __T0 = 100e-6
-            __tExp = 3.3
-            __tResp = 0e-6  # no waiting 
-            self.dev.write(':PULSE0:PER {}'.format(__T0))  # 0.000100080   
-            ## Channel A: Magnetic Field 
-            self.__exp(__T0,1,'INV',__tExp,3.0,0.0,1,0) 
-            ## Channel B: Repumper
-            if re.search('(repump)',cfg,re.IGNORECASE) != None:   # leave repump on
-                __repumpTime = 3.0 + 0.1
-            else:
-                __repumpTime = 3.0
-            self.__exp(__T0,2,'NORM',__tExp,__repumpTime,0.0,1,0)
-            ## Channel E: Cooling
-            if re.search('(cool)',cfg,re.IGNORECASE) != None:   # leave cooling on
-                __coolTime = 3.0 + 0.1
-            else:
-                __coolTime = 3.0
-            self.__exp(__T0,5,'INV',__tExp,__coolTime,0.0,1,0) 
-            ## added for cooling
-            self.__exp(__T0,6,'INV',__tExp,0.1,0.0,1,30001) 
-            self.mux(5,5,6)
-            ## Channel H: TTL for quTAU
-            self.__exp(__T0,8,'NORM',__tExp,0.000050,__tResp,1000,30000)
-            self.on(1,2,5,6,8)
+            
     @sleep_method
     def burst(self, ch, n_pulses):
-        self.lev(ch)
-        self.pol(ch,'NORM')
-        self.on(ch)
-        self.trigOff()
         self.dev.write(':PULSe{}:CMOD BURS'.format(ch))
-        time.sleep(self.t_sleep)
         self.dev.write(':PULSe{}:BCOunter {}'.format(ch,n_pulses))
 
 #================================================================
@@ -758,12 +788,12 @@ class Agilent_ESG_SG:
         #self.dev.write(':FREQuency {} {}'.format(f,unit)) #for general units
         self.dev.write(':FREQuency {} MHz'.format(f))
 
-    def lev(self,amp):  #define ampitude and unit {dBm,dBUV,V,VEMF}
+    def lev(self,amplitude):  #define ampitude and unit {dBm,dBUV,V,VEMF}
         #self.dev.write('POWer:AMPLitude {} {}'.format(amp,unit))
-        self.dev.write('POWer:AMPLitude {} dBm'.format(amp))
+        self.dev.write('POWer:AMPLitude {} dBm'.format(amplitude))
         
-    def offset(self,amp,unit):  #define offset and unit {dBm,dBUV,V,VEMF}
-        self.dev.write(':POWer:OFFSet {} {}'.format(amp,unit))
+    def offset(self,amplitude,unit):  #define offset and unit {dBm,dBUV,V,VEMF}
+        self.dev.write(':POWer:OFFSet {} {}'.format(amplitude,unit))
 
     def phase(self,phase,unit): #define phase and unit {radian,degrees}
         self.dev.write(':PHASe {} {}'.format(phase,unit))
@@ -772,15 +802,11 @@ class Agilent_ESG_SG:
 # Tektronix AFG3000 Series Arbitrary Function Generator
 #================================================================
 
-class Tektronix_AFG3000:
+class tektronix_AFG3000:
     def __init__(self,address,*arg):
-        if arg:
-            self.address = address #'TCPIP::<IP ADDRESS>::INSTR'
-            self.dev = vxi11.Instrument(self.address)
-        else:
-            self.address = address 
-            self.rm = pyvisa.ResourceManager()
-            self.dev = self.rm.open_resource(self.address)
+        self.address = address 
+        self.rm = pyvisa.ResourceManager()
+        self.dev = self.rm.open_resource(self.address)
         
     def reset(self): #Reset
         self.dev.write("*RST")
@@ -791,10 +817,17 @@ class Tektronix_AFG3000:
     def on(self): # default: both 
         self.dev.write(':OUTPut ON')
 
-    def freq(self,ch,f):
-        self.dev.write(':SOURCe{}:Freq:FIXed {}'.format(ch,self.__Hz(f)));
+    def freq(self,f):
+        self.dev.write('SOURce:FREQuency:FIXed {}'.format(self.__Hz(f)));
 
-    def __Hz(self,f):  # in Hz, support unit. Default: MHz
+    def __Hz(f): 
+        """
+        Description: Sets all frequencies to MHz Unit
+
+        Input: f : frequency : float
+
+        Output: f * 1e6 : frequency in MHz : float
+        """
         if type(f) == str:
             if re.search('[mM]',f) != None:
                 return 1e6*float(re.sub('[a-zA-Z]','',f))
@@ -805,25 +838,28 @@ class Tektronix_AFG3000:
         else: # float, or str that contains only numbers
             return float(f)*1e6
 
-    def lev(self,ch,v):
+    def lev(self,v):
         if type(v) == str:
             __V = float(re.sub('[a-zA-Z]','',v))  # unitless value
             if re.search('[rR]',v) != None:  #  VRMS
-                self.dev.write(':SOURCe{}:VOLTage:UNIT VRMS'.format(ch))
+                self.dev.write('SOURCe:VOLTage:UNIT VRMS')
             elif re.search('[dD]',v) != None:  # dBm
-                self.dev.write(':SOURCe{}:VOLTage:UNIT DBM'.format(ch))
+                self.dev.write('SOURCe:VOLTage:UNIT DBM')
         else:  # default: [Vpp] 
             __V = v
-            self.dev.write(':SOURCe{}:VOLTage:UNIT VPP'.format(ch))
+            self.dev.write(':SOURCe:VOLTage:UNIT VPP')
 
-        self.dev.write(':SOURCe{}:VOLTage {}'.format(ch,__V))
-        self.dev.write(':SOURCe{}:VOLTage:UNIT VPP'.format(ch))
+        self.dev.write('SOURCe:VOLTage {}'.format(__V))
+        self.dev.write('SOURCe:VOLTage:UNIT VPP')
 
-    def offset(self,ch,o):  # V_DC
-        self.dev.write(':SOURCe{}:VOLTage:OFFSet {}'.format(ch,o));
+    def offset(self,offset):  # V_DC
+        self.dev.write('SOURC:VOLTage:OFFSet {}'.format(offset));
 
-    def phase(self,ch,p):
-        self.dev.write(':SOURCe{}:PHASe {}'.format(ch,p));
+    def phase(self,phase):
+        self.dev.write('SOURC:PHASe {}'.format(phase));
+        
+    def burst_delay(self,tdelay):
+        self.dev.write('SOUR:BURS:TDEL {}ns'.format(tdelay))
         
     def gaussian(self,t,mu,FWHM,a): #Gaussian Function. Inputs: (FWHM, Amplitude, Center)
         sigma = (FWHM)/(2*np.sqrt(2*np.log(2)))
@@ -833,55 +869,242 @@ class Tektronix_AFG3000:
     def square(self,t,leadingedge,width,amp): #square pulse with duty cycle
         return np.piecewise(t,[(t<=leadingedge),((t>leadingedge) & (t<leadingedge+width)),(t>=leadingedge+width)],[0,amp,0])
     
-    def arb(self,ch,ν,func,*arg):
-        total_time = 1/(self.__Hz(ν))
-        
-        if inspect.ismethod(func) == True:
-            t = np.linspace(0,total_time,1000)
-            data = func(t,*arg)
-            datastring = ",".join(map(str,data))
-        else:
-            data = func
-            datastring = ",".join(map(str,data))
-        
-        self.dev.write("SOURCE{}:BURST OFF".format(ch))
-        self.dev.write("OUTPUT{} ON".format(ch))
-        self.dev.write("SOURCE{}:TRACE:DATA VOLATILE,".format(ch)+ datastring)
-        self.dev.write("SOURCE{}:VOLTAGE:UNIT VRMS".format(ch))
-        self.dev.write('SOURCE{}:Freq {}'.format(ch,self.__Hz(ν)))
-        self.dev.write("SOURCE{}:VOLTAGE:LOW {}".format(ch,min(data)))
-        self.dev.write("SOURCE{}:VOLTAGE:HIGH {}".format(ch,max(data)*2))
-        self.dev.write("SOURCE{}:VOLTAGE:OFFSET 0".format(ch))
-        self.dev.write("SOURCE{}:PHASE 0".format(ch))
-        self.dev.write("SOURCE{}:PERIOD {}".format(ch,total_time))
-        self.dev.write("SOURCE{}:PHASE:INIT".format(ch))
+    def normalize(self, waveform):
+        """
+        Description: Normalizes data for arbitrary waveform design, points are limited
+        to -1 to 1 Volt
+
+        Input: data : waveform : np.array or list
+
+        Output: np.array(waveform)/np.absolute(max(waveform)) : normalized data :  np.array or list
+        """
+        return np.array(waveform)/np.absolute(max(waveform))
     
+    def arbitrary(self, signal_width, waveform, *arg):
+        """
+        Description: Allows one to set and create arbitrary waveform output 
+
+        Input: signal_width : width (time) of the argument in seconds : float
+               waveform : method or list of values : method or array
+               arg* : arguments of the waveform function : misc
+
+        Output: None : class method
+        """
+        buffer_size = 2**14-2
+        
+        if data[0] != data[-1]:
+            data[0] = 0
+            data[-1] = data[0]
+        else:
+            pass
+        
+        if inspect.ismethod(waveform) == True:
+
+            t = np.linspace(0,total_time,buffer_size)
+            data = waveform(t,*arg)
+
+            datastring = self.normalize(data)
+            m = buffer_size / (datastring.max() - datastring.min())
+            b = -m * datastring.min()
+            dac_values = (m * datastring + b)
+            np.around(dac_values, out=dac_values)
+            dac_values = dac_values.astype(np.uint16)
+
+        else:
+            data = waveform
+            datastring = self.normalize(data)
+            m = buffer_size / (datastring.max() - datastring.min())
+            b = -m * datastring.min()
+            dac_values = (m * datastring + b)
+            np.around(dac_values, out=dac_values)
+            dac_values = dac_values.astype(np.uint16)    
+
+        self.dev.write('DATA:DEFine EMEMory,{}'.format(len(data)))
+        self.dev.write_binary_values("DATA:DATA EMEM1,", dac_values, datatype="H", is_big_endian=True)
+        self.dev.write("SOURce:FUNC:SHAPE EMEM1")
+        self.dev.write("SOURce1:VOLTage:LEVel:IMMediate:LOW {}".format(min(data)))
+        self.dev.write("SOURce1:VOLTage:LEVel:IMMediate:HIGH {}".format(max(data)))
+        self.dev.write("SOURCE:FREQ {}".format(self.__Hz(freq)))
+
     def ext_trig(self):
         self.dev.write("TRIGger:SEQuence:SOURce EXTernal")
     
-    def arb_burst(self,ch,ν,cycles,func,*arg):
-        total_time = 1/(self.__Hz(ν))
+    def burst(self, ch, mode, cycles):
+        """
+        Description: Allows on to use burst functionallity
+
+        Input: ch : channel : int
+               mode : burst mode : int = {'TRIG' = 0, 'GAT' = 1,'INF' = 2}
+               cycles : cycles : int
+
+        Output: 
+        """
+        modes = ['TRIGgered','GATed','INFinity']
+        self.dev.write('SOURce{}:BURSt:MODE {}'.format(ch,modes[mode]))
+        self.dev.write('SOURce{}:BURSt:TRIGger:SOURce EXT'.format(ch))
+        self.dev.write('SOURce{}:BURSt:NCYCles {}'.format(ch,cycles))
+        self.dev.write('SOURce{}:BURSt ON'.format(ch))
+    
+    def DC(self, offset):
+        """
+        Description: Enables DC Mode (Tested 04/03/2022)
+
+        Input: ch : channel : int
+               offset : DC Offset : float
+
+        Output: None : class method
+        """
+        self.dev.write("SOUR:FUNC:SHAP DC")
+        self.dev.write("SOUR:VOLT:LEV:IMM:OFFS {}".format(offset))
         
-        if inspect.ismethod(func) == True:
-            t = np.linspace(0,total_time,1000)
-            data = func(t,*arg)
-            datastring = ",".join(map(str,data))
+        
+
+#================================================================
+# Rigol DP8000 Series DC Power Supply
+#================================================================        
+
+class Rigol_DP800:
+
+    def __init__(self,address): 
+        """ 
+        Description: This function will initialize the power supply class object 
+    
+        Input: IP address : string
+        Output: power supply object with built in funtions : class object
+
+        Example: 
+        >>gen1 = gen.Rigol_DP800('TCPIP::10.2.1.158')
+
+        """
+        self.address = address            
+        self.rm = pyvisa.ResourceManager()
+        self.dev = self.rm.open_resource(self.address)
+        
+    def inspect(self): 
+        """ 
+        Description: This function will return to you the current settings of the 
+        power supply. This includes overvoltage/overcurrent protection, overvoltage/current protection value 
+        state on each channel, range of each channel
+    
+    
+        Input: None : None
+        Output: Characterization List : String
+
+        Example: 
+        >>gen1.inspect()
+        
+        (CH1 :: Mode : CV :: OCP state : ON :: OCP value : 0.5mA ::  OVP state : OFF :: OVP value : 0.5V :: Range : 10A/5V :: State : ON)
+        etc...
+
+        """
+        try: 
+            for i in range(1,4):
+                ch = 'CH{} :: '.format(i)
+                ch = ch,'Mode : ',str(self.dev.query(':OUTP:MODE? CH{};*OPC?'.format(i))),' :: '
+                ch = ch,'OCP state : ',str(self.dev.query(':OUTP:OCP? CH{};*OPC?'.format(i))),' :: '
+                ch = ch,'OCP value : ',str(self.dev.query(':OUTP:OCP:VAL? CH{};*OPC?'.format(i))),' :: '
+                ch = ch,'OVP state : ',str(self.dev.query(':OUTP:OVP? CH{};*OPC?'.format(i))),' :: '
+                ch = ch,'OVP value : ',str(self.dev.query(':OUTP:OVP:VAL? CH{};*OPC?'.format(i))),' :: '
+                print(ch)
+        except:
+            print('unable to obtain parameters')
+        return
+        
+    def set_safety(self): 
+        """ 
+        Description: This function will return to you the current settings of the 
+        power supply. This includes overvoltage/overcurrent protection, overvoltage/current protection value 
+        state on each channel, range of each channel
+    
+    
+        Input: None : None
+        Output: Characterization List : String
+
+        Example: 
+        >>gen1.set_safety()
+
+        """
+        try: 
+            for i in range(1,4):
+                self.dev.write(':OUTP:OCP CH{},ON'.format(i))
+                self.dev.write(':OUTP:OVP CH{},ON'.format(i))
+            print('SAFETY SET')
+        except:
+            print('unable to excecute')
+        return
+    
+    def on(self, *ch): 
+        """ 
+        Description: Will turn on specified channel
+    
+    
+        Input: ch : [1,2,3]
+        Output: Method which turns the output state to on
+
+        Example: 
+        >>gen1.on(1)
+
+        """
+        if bool(ch) == True:
+            self.dev.write(':OUTP CH{}, ON'.format(ch[0]))
         else:
-            data = func
-            datastring = ",".join(map(str,data))
-        
-        self.dev.write("OUTPUT{} ON".format(ch))
-        self.dev.write("SOURCE{}:TRACE:DATA VOLATILE,".format(ch)+ datastring)
-        self.dev.write("SOURCE{}:VOLTAGE:UNIT VRMS".format(ch))
-        self.dev.write('SOURCE{}:Freq {}'.format(ch,self.__Hz(ν)))
-        self.dev.write("SOURCE{}:VOLTAGE:LOW {}".format(ch,min(data)))
-        self.dev.write("SOURCE{}:VOLTAGE:HIGH {}".format(ch,max(data)*2))
-        self.dev.write("SOURCE{}:VOLTAGE:OFFSET 0".format(ch))
-        self.dev.write("SOURCE{}:PHASE 0".format(ch))
-        self.dev.write("SOURCE{}:PERIOD {}".format(ch,total_time))
-        self.dev.write("SOURCE{}:PHASE:SYNC".format(ch))
-        
-        #triggered burst
-        self.dev.write("SOURCE{}:BURST ON".format(ch))
-        self.dev.write("SOURCE{}:BURST:NCYC {}".format(ch,cycles))
-        self.dev.write("SOURCE{}:BURST:MODE:TRIG".format(ch))
\ No newline at end of file
+            for i in range(1,4):
+                self.dev.write(':OUTP CH{}, ON'.format(i))
+        return
+    
+    def off(self, *ch): 
+        """ 
+        Description: Will turn off specified channel
+    
+    
+        Input: ch : [1,2,3]
+        Output: Method which turns the output state to off
+
+        Example: 
+        >>gen1.off(2)
+
+        """
+        if bool(ch) == True:
+            self.dev.write(':OUTP CH{}, OFF'.format(ch[0]))
+        else:
+            for i in range(1,4):
+                self.dev.write(':OUTP CH{}, OFF'.format(i))
+        return
+    
+    def current(self, ch, *current): 
+        """ 
+        Description: This function will set or get the current of the specified channel
+    
+    
+        Input: channel : [1,2,3], current : float
+        Output: Method which changes the current
+
+        Example: 
+        >>gen1.current(1,2)
+
+        """
+        if current:
+            self.dev.write('SOUR{}:CURR {}'.format(ch,current[0])) #Units Amps
+        else:
+            self.dev.write('SOUR{}:CURR?'.format(ch))
+ 
+        return
+
+    def voltage(self, ch, *voltage): 
+        """ 
+        Description: This function will set or get the voltage of the specified channel
+    
+    
+        Input: channel : [1,2,3], voltage : float
+        Output: Method which changes the voltage
+
+        Example: 
+        >>gen1.voltage(1,2)
+
+        """
+        if bool(voltage) == True:
+            self.dev.write('SOUR{}:VOLT {}'.format(ch,voltage[0])) #Units Volts
+        else:
+            print(self.dev.query('SOUR{}:VOLT?'.format(ch)))
+ 
+        return
\ No newline at end of file
